@@ -4,15 +4,19 @@ use strict;
 use warnings;
 use integer;
 use bytes;
-use version; our $VERSION = version->declare('v0.0.2');
+use version; our $VERSION = version->declare('v0.0.4');
 
+use feature 'state';
+
+use GDPR::IAB::TCFv2::BitUtils qw(:all);
+use GDPR::IAB::TCFv2::BitField;
+use GDPR::IAB::TCFv2::RangeSection;
 use MIME::Base64 qw<decode_base64>;
 use Carp         qw<croak>;
 
 sub CONSENT_STRING_TCF2_SEPARATOR {'.'}
 sub CONSENT_STRING_TCF2_PREFIX    {'C'}
 sub MIN_BYTE_SIZE                 {29}
-sub MIN_BIT_SIZE                  { 8 * MIN_BYTE_SIZE }
 
 # ABSTRACT: gdpr iab tcf v2 consent string parser
 
@@ -23,10 +27,10 @@ sub Parse {
 
     my $core_tc_string = _get_core_tc_string($tc_string);
 
-    my $data = unpack 'B*', decode_base64($core_tc_string);
+    my $data = unpack 'B*', _decode_base64url($core_tc_string);
 
     croak "vendor consent strings are at least @{[ MIN_BYTE_SIZE ]} bytes long"
-      if length($data) < MIN_BIT_SIZE;
+      if length($data) / 8 < MIN_BYTE_SIZE;
 
     my $self = {
         data           => $data,
@@ -40,6 +44,52 @@ sub Parse {
 
     croak 'invalid vendor list version' if $self->vendor_list_version == 0;
 
+    my $vendor_consents;
+    my $legitimate_interest_start;
+
+    if ( $self->_is_vendor_consent_range_encoding ) {
+        ( $vendor_consents, $legitimate_interest_start ) =
+          $self->_parseRangeSection( $self->max_vendor_id_consent, 230 );
+    }
+    else {
+        ( $vendor_consents, $legitimate_interest_start ) =
+          $self->_parseBitField( $self->max_vendor_id_consent, 230 );
+    }
+
+    $self->{vendor_consents} = $vendor_consents;
+
+    my $legitimate_interest_max_vendor =
+      get_uint16( $self->{data}, $legitimate_interest_start );
+
+    $self->{legitimate_interest_max_vendor} = $legitimate_interest_max_vendor;
+
+    croak
+      "invalid consent data: no legitimate interest start position (got $legitimate_interest_start +16 but @{[ length( $self->{data} ) ]})"
+      if $legitimate_interest_start + 16 > length( $self->{data} );
+
+    my $is_vendor_legitimate_interest_range =
+      is_set( $data, $legitimate_interest_start + 16 );
+
+    my $vendor_legitimate_interests;
+    my $pub_restrict_start;
+
+    if ($is_vendor_legitimate_interest_range) {
+        ( $vendor_legitimate_interests, $pub_restrict_start ) =
+          $self->_parseRangeSection(
+            $self->max_vendor_id_legitimate_interest,
+            $legitimate_interest_start + 17
+          );
+    }
+    else {
+        ( $vendor_legitimate_interests, $pub_restrict_start ) =
+          $self->_parseBitField(
+            $self->max_vendor_id_legitimate_interest,
+            $legitimate_interest_start + 17
+          );
+    }
+
+    $self->{vendor_legitimate_interests} = $vendor_legitimate_interests;
+
     return $self;
 }
 
@@ -48,19 +98,28 @@ sub _get_core_tc_string {
 
     my $pos = index( $tc_string, CONSENT_STRING_TCF2_SEPARATOR );
 
+    return $tc_string if $pos < 0;
+
     return substr( $tc_string, 0, $pos );
 }
 
+sub _decode_base64url {
+    my $s = shift;
+    $s =~ tr[-_][+/];
+    $s .= '=' while length($s) % 4;
+    return decode_base64($s);
+}
+
 sub version {
     my $self = shift;
 
-    return $self->_get_uint8( 0, 6 );
+    return get_uint6( $self->{data}, 0 );
 }
 
 sub created {
     my $self = shift;
 
-    my $deciseconds = $self->_get_uint64( 6, 36 );
+    my $deciseconds = get_uint36( $self->{data}, 6 );
 
     return $deciseconds / 10;
 }
@@ -68,7 +127,7 @@ sub created {
 sub last_updated {
     my $self = shift;
 
-    my $deciseconds = $self->_get_uint64( 42, 36 );
+    my $deciseconds = get_uint36( $self->{data}, 42 );
 
     return $deciseconds / 10;
 }
@@ -76,34 +135,58 @@ sub last_updated {
 sub cmp_id {
     my $self = shift;
 
-    return $self->_get_uint16( 78, 12 );
+    return get_uint12( $self->{data}, 78 );
 }
 
 sub cmp_version {
     my $self = shift;
 
-    return $self->_get_uint16( 90, 12 );
+    return get_uint12( $self->{data}, 90 );
 }
 
 sub consent_screen {
     my $self = shift;
 
-    return $self->_get_uint8( 102, 6 );
+    return get_uint6( $self->{data}, 102 );
 }
 
 sub consent_language {
     my $self = shift;
 
-    return join "", map { chr( $_ + 65 ) } (
-        $self->_get_uint8( 108, 6 ),
-        $self->_get_uint8( 114, 6 ),
-    );
+    return get_char6_sequence( $self->{data}, 108, 2 );
 }
 
 sub vendor_list_version {
     my $self = shift;
 
-    $self->_get_uint16( 120, 12 );
+    return get_uint12( $self->{data}, 120 );
+}
+
+sub policy_version {
+    my $self = shift;
+
+    return get_uint6( $self->{data}, 132 );
+}
+
+sub is_service_specific {
+    my $self = shift;
+
+    return is_set( $self->{data}, 138 );
+}
+
+sub use_non_standard_stacks {
+    my $self = shift;
+
+    return is_set( $self->{data}, 139 );
+}
+
+sub is_special_feature_opt_in {
+    my ( $self, $id ) = @_;
+
+    croak "invalid special feature id $id: must be between 1 and 12"
+      if $id < 1 || $id > 12;
+
+    return is_set( $self->{data}, 140 + $id - 1 );
 }
 
 sub is_purpose_consent_allowed {
@@ -112,46 +195,82 @@ sub is_purpose_consent_allowed {
     croak "invalid purpose id $id: must be between 1 and 24"
       if $id < 1 || $id > 24;
 
-    return $self->_is_set( 152 + $id - 1 );
+    return is_set( $self->{data}, 152 + $id - 1 );
 }
 
-sub _is_set {
-    my ( $self, $offset ) = @_;
+sub is_purpose_legitimate_interest_allowed {
+    my ( $self, $id ) = @_;
 
-    return substr( $self->{data}, $offset, 1 ) == 1;
+    croak "invalid purpose id $id: must be between 1 and 24"
+      if $id < 1 || $id > 24;
+
+    return is_set( $self->{data}, 176 + $id - 1 );
 }
 
-sub _get_uint8 {
-    my ( $self, $offset, $nbits ) = @_;
+sub purpose_one_treatment {
+    my $self = shift;
 
-    my $padding = "0" x ( 8 - $nbits );
+    return is_set( $self->{data}, 200 );
+}
 
-    return unpack(
-        "C",
-        pack( "B8", $padding . substr( $self->{data}, $offset, $nbits ) )
+sub publisher_country_code {
+    my $self = shift;
+
+    return get_char6_sequence( $self->{data}, 201, 2 );
+}
+
+sub max_vendor_id_consent {
+    my $self = shift;
+
+    return get_uint16( $self->{data}, 213 );
+}
+
+sub max_vendor_id_legitimate_interest {
+    my $self = shift;
+
+    return $self->{legitimate_interest_max_vendor};
+}
+
+sub vendor_consent {
+    my ( $self, $id ) = @_;
+
+    return $self->{vendor_consents}->contains($id);
+}
+
+sub vendor_legitimate_interest {
+    my ( $self, $id ) = @_;
+
+    return $self->{vendor_legitimate_interests}->contains($id);
+}
+
+sub _is_vendor_consent_range_encoding {
+    my $self = shift;
+
+    return is_set( $self->{data}, 229 );
+}
+
+sub _parseRangeSection {
+    my ( $self, $vendor_bits_required, $start_bit ) = @_;
+
+    my $range_section = GDPR::IAB::TCFv2::RangeSection->new(
+        data                 => $self->{data},
+        start_bit            => $start_bit,
+        vendor_bits_required => $vendor_bits_required,
     );
+
+    return ( $range_section, $range_section->current_offset );
 }
 
-sub _get_uint16 {
-    my ( $self, $offset, $nbits ) = @_;
+sub _parseBitField {
+    my ( $self, $vendor_bits_required, $start_bit ) = @_;
 
-    my $padding = "0" x ( 16 - $nbits );
-
-    return unpack(
-        "S>",
-        pack( "B16", $padding . substr( $self->{data}, $offset, $nbits ) )
+    my $bitfield = GDPR::IAB::TCFv2::BitField->new(
+        data                 => $self->{data},
+        start_bit            => $start_bit,
+        vendor_bits_required => $vendor_bits_required,
     );
-}
 
-sub _get_uint64 {
-    my ( $self, $offset, $nbits ) = @_;
-
-    my $padding = "0" x ( 64 - $nbits );
-
-    return unpack(
-        "Q>",
-        pack( "B64", $padding . substr( $self->{data}, $offset, $nbits ) )
-    );
+    return ( $bitfield, $start_bit + $vendor_bits_required );
 }
 
 sub looksLikeIsConsentVersion2 {
@@ -169,7 +288,7 @@ GDPR::IAB::TCFv2 - Transparency & Consent String version 2 parser
 
 =head1 VERSION
 
-Version v0.0.2
+Version v0.0.4
 
 =head1 SYNOPSIS
 
@@ -199,6 +318,14 @@ The purpose of this package is to parse Transparency & Consent String (TC String
         $consent->is_purpose_consent_allowed($_)
     } (1, 3, 9, 10);
 
+    say "weborama (vendor id 284) has consent" if $consent->vendor_consent(284);
+
+=head1 ACRONYMS
+
+GDPR: General Data Protection Regulation L<https://iabeurope.eu/about-us/>
+IAB: Interactive Advertising Bureau L<About IAB|https://iabeurope.eu/about-us/>
+TCF: The Transparency & Consent Framework L<TCF v2.2|https://iabeurope.eu/transparency-consent-framework/>
+
 =head1 CONSTRUCTOR
 
 =head2 Parse
@@ -255,6 +382,47 @@ Version of the GVL used to create this TC String.
 
 The user's consent value for each Purpose established on the legal basis of consent.
 
+    my $ok = $instance->is_purpose_consent_allowed(1);
+
+=head2 is_purpose_legitimate_interest_allowed
+
+The user's consent value for each Purpose established on the legal basis of legitimate interest.
+
+    my $ok = $instance->is_purpose_legitimate_interest_allowed(1);
+
+=head2 purpose_one_treatment
+
+CMPs can use the PublisherCC field to indicate the legal jurisdiction the publisher is under to help vendors determine whether the vendor needs consent for Purpose 1.
+
+Returns true if Purpose 1 was NOT disclosed at all.
+
+Returns false if Purpose 1 was disclosed commonly as consent as expected by the L<Policies|https://iabeurope.eu/iab-europe-transparency-consent-framework-policies/>.
+
+=head2 publisher_country_code
+
+Two-letter L<ISO 639-1|https://en.wikipedia.org/wiki/ISO_639-1> language code of the country that determines legislation of reference. 
+Commonly, this corresponds to the country in which the publisher's business entity is established.
+
+=head2 max_vendor_id_consent
+
+The maximum Vendor ID that is represented in the following bit field or range encoding.
+
+Because this section can be a variable length, this indicates the last ID of the section so that a decoder will know when it has reached the end.
+
+=head2 vendor_consent
+
+The consent value for each Vendor ID 
+
+=head2 max_vendor_id_legitimate_interest
+
+The maximum Vendor ID that is represented in the following bit field or range encoding.
+
+Because this section can be a variable length, this indicates the last ID of the section so that a decoder will know when it has reached the end.
+
+=head2 vendor_legitimate_interest
+	
+The legitimate interest value for each Vendor ID
+
 =head1 FUNCTIONS
 
 =head2 looksLikeIsConsentVersion2
